
__all__ = ["regplot"]

import random
from typing import Union, List

import numpy as np
import matplotlib.pyplot as plt
from easy_mpl.utils import process_axes
from easy_mpl.utils import AddMarginalPlots


REGC_COMBS = [
    ['cadetblue', 'slateblue', 'darkslateblue'],
    ['cadetblue', 'mediumblue', 'mediumblue'],
    ['cornflowerblue', 'dodgerblue', 'darkblue'],
    ['cornflowerblue', 'dodgerblue', 'steelblue'],
    ['cornflowerblue', 'mediumblue', 'dodgerblue'],
    ['cornflowerblue', 'steelblue', 'mediumblue'],
    ['darkslateblue', 'aliceblue', 'mediumblue'],
    ['darkslateblue', 'blue', 'royalblue'],
    ['darkslateblue', 'blueviolet', 'royalblue'],
    ['darkslateblue', 'darkblue', 'midnightblue'],
    ['darkslateblue', 'mediumblue', 'darkslateblue'],
    ['darkslateblue', 'midnightblue', 'mediumblue'],
    ['seagreen', 'darkslateblue', 'cadetblue'],
    ['cadetblue', 'darkblue', 'midnightblue'],
    ['cadetblue', 'deepskyblue', 'cadetblue']
]


def regplot(
        x: Union[np.ndarray,  list],
        y: Union[np.ndarray, list],
        label:str = None,
        marker_size: Union[int, float] = 20,
        marker_color=None,
        scatter_kws: dict = None,
        line_style='-',
        line_color=None,
        line_kws: dict = None,
        ci: Union[int, None] = 95,
        fill_color=None,
        figsize: tuple = None,
        marginals: bool = False,
        marginal_ax_size: Union[float, List[float]] = 0.7,
        marginal_ax_pad: Union[float, List[float]] = 0.25,
        ridge_line_kws: Union[dict, List[dict]] = None,
        fill_kws: Union[dict, List[dict]] = None,
        hist: bool = True,
        hist_kws: Union[dict, List[dict]] = None,
        ax: plt.Axes = None,
        ax_kws:dict = None,
        show: bool = True,
) -> plt.Axes:
    """
    Regpression plot with regression line and confidence interval

    Parameters
    ----------
        x : array like, optional
            the 'x' value. It can be numpy array, pandas DataFram/Series or a list
        y : array like, optional
             It can be numpy array, pandas DataFram/Series or a list
        ci : int, optional
            confidence interval. Set to None if not required.
        label : str, optional
            Label to use as legend.
            The value to annotate with.
        marker_size : int, optional
            size of marker
        marker_color: optional
            color of marker
        scatter_kws : dict
            keyword arguments for :obj:`matplotlib.axes.scatter`
        line_style : str (default='-')
            line style will be used as ax.plot(x,y,line_style)
        line_color : optional
            color of line
        line_kws : dict
            keyword arguments for axes.plot for line plot
        fill_color : optional
            only relevent if ci is not None.
        figsize : tuple, optional
            figure size (width, height)
        marginals : bool (default=False)
            whether to draw the marginal plots or not. If yes, marginal plots
            will
            be drawn using :py:class:`easy_mpl.utils.AddMarginalPlots`
        marginal_ax_size :
            size of marginal axes. If given as list, the first argument is taken
            for
            horizontal marginal axes and second argument for vertical merginal
            axees.
            It is only valid if ``marginals`` is set to True.
        marginal_ax_pad :
            pad value for marginal axes. If given as list, the first argument
            is taken for
            horizontal marginal axes and second argument for vertical merginal
            axees.
            It is only valid if ``marginals`` is set to True.
        ridge_line_kws :
            keyword arguments for a:obj:`matplotlib.axes.Axes.plot to draw the
             kde.
            It can be dictionary
            or list of two dictionaries, where first dictionary is for hoziontal
            marginal axes and second dictionary is for vertical marginal axes
        fill_kws : dict/List[dict] (default=None)
            keyword arguments for :obj:`matplotlib.axes.Axes.fill_between`. It
            can be dictionary or list of
            two dictionaries, where first dictionary is for hoziontal
            marginal axes and second dictionary is for vertical marginal axes
        hist : bool
            whether to draw the histogram or not on marginal axes
        hist_kws :
            keyword arguments for :obj:`matplotlib.axes.Axes.hist`. It can be
            dictionary or list of
            two dictionaries, where first dictionary is for hoziontal
            marginal axes and second dictionary is for vertical marginal axes
        ax : plt.Axes, optional
            matplotlib axes :obj:`matplotlib.axes` to draw plot on. If not given,
            current avaialable will be used.
        ax_kws : dict (default=None)
            keyword arguments for :py:func:`easy_mpl.utils.process_axes`
        show : bool, optional
            whether to show the plot or not

    Returns
    --------
    matplotlib.pyplot.Axes
        the matplotlib Axes :obj:`matplotlib.axes` on which regression plot is
        drawn.

    Examples
    --------
        >>> import numpy as np
        >>> from easy_mpl import regplot
        >>> x_, y_ = np.random.random(100), np.random.random(100)
        >>> regplot(x_, y_)

    See :ref:`sphx_glr_auto_examples_reg_plot.py` for more examples

    Note
    ----
        If nans are present in x or y, they will be removed.

    """
    x = np.array(x).reshape(-1,)
    y = np.array(y).reshape(-1,)

    # removing nans based upon nans in x
    x_nan_idx = np.isnan(x)
    if x_nan_idx.sum() > 0:
        x = x[~x_nan_idx]
        y = y[~x_nan_idx]

    # removing nans based upon nans in y
    y_nan_idx = np.isnan(y)
    if y_nan_idx.sum() > 0:
        x = x[~y_nan_idx]
        y = y[~y_nan_idx]

    assert len(x) > 1, f"""
    length of x is smaller than 1 {x} 
    {len(y_nan_idx)} nans found in y and   {len(x_nan_idx)} nans found in x"""
    assert len(x) == len(y), f"""
    x and y must be same length. Got {len(x)} and {len(y)}"""

    mc, lc, fc = random.choice(REGC_COMBS)
    _metric_names = {'r2': '$R^2$'}

    if ax is None:
        _, ax = plt.subplots(figsize=figsize or (6, 5))

    _scatter_kws = dict()
    if label:
        _scatter_kws['label'] = label

    if scatter_kws is not None:
        _scatter_kws.update(scatter_kws)

    if marker_color is None:
        marker_color = mc

    ax.scatter(x, y, c=marker_color,
               s=marker_size, **_scatter_kws)

    if label:
        ax.legend()

    if line_kws is None:
        line_kws = dict()

    if line_color is None:
        line_color = lc

    if fill_color is None:
        fill_color = fc

    _regplot(x,
             y,
             ax=ax,
             ci=ci,
             line_style=line_style,
             line_color=line_color,
             fill_color=fill_color,
             **line_kws)

    if marginals:
        AddMarginalPlots(ax,
                         pad=marginal_ax_pad,
                         size=marginal_ax_size,
                         hist=hist,
                         hist_kws=hist_kws,
                         ridge_line_kws=ridge_line_kws,
                         fill_kws=fill_kws)(x, y)

    _ax_kws = {'xlabel': 'Observed',
               'xlabel_kws': {'fontsize': 14},
               'ylabel': 'Prediction',
               'ylabel_kws': {'fontsize': 14},
               }

    if ax_kws is not None:
        _ax_kws.update(ax_kws)

    process_axes(ax, **_ax_kws)

    if show:
        plt.show()

    return ax


def bootdist(f, args, n_boot=1000, **func_kwargs):

    n = len(args[0])
    integers = np.random.randint
    boot_dist = []
    for i in range(int(n_boot)):
        resampler = integers(0, n, n, dtype=np.intp)  # intp is indexing dtype
        sample = [a.take(resampler, axis=0) for a in args]
        boot_dist.append(f(*sample, **func_kwargs))

    return np.array(boot_dist)


def _regplot_paras(x, y, ci: int = None):
    """prepares parameters for regplot"""
    grid = np.linspace(np.min(x), np.max(x), 100)
    x = np.c_[np.ones(len(x)), x]
    grid = np.c_[np.ones(len(grid)), grid]
    yhat = grid.dot(reg_func(x, y))

    err_bands = None
    if ci:
        boots = bootdist(reg_func, args=[x, y], n_boot=1000).T

        yhat_boots = grid.dot(boots).T
        err_bands = _ci(yhat_boots, ci, axis=0)

    return grid, yhat, err_bands


def _ci(a, which=95, axis=None):
    """Return a percentile range from an array of values."""
    p = 50 - which / 2, 50 + which / 2
    return np.nanpercentile(a, p, axis)


def reg_func(_x, _y):
    return np.linalg.pinv(_x).dot(_y)


def _regplot(x, y, ax, ci=None,
             line_style="-",
             line_color=None, fill_color=None, **kwargs):

    grid, yhat, err_bands = _regplot_paras(x, y, ci)

    ax.plot(grid[:, 1], yhat, line_style, color=line_color, **kwargs)

    if ci:
        ax.fill_between(grid[:, 1], *err_bands,
                        facecolor=fill_color,
                        alpha=.15,
                        where=np.array([True for _ in range(len(grid))])
                        )
    return ax




__all__ = ["boxplot"]

from typing import Union, List, Tuple

import matplotlib
import numpy as np
import matplotlib.pyplot as plt

from .utils import is_rgb
from .utils import is_series
from .utils import is_dataframe
from .utils import process_axes
from .utils import create_subplots
from .utils import make_cols_from_cmap


def boxplot(
        data:Union[np.ndarray, List[np.ndarray]],
        line_color:Union[str, List[str]] = None,
        line_width = None,
        fill_color:Union[str, List[str]] = None,
        labels:Union[str, List[str]] = None,
        share_axes:bool = True,
        figsize:tuple = None,
        ax:plt.Axes = None,
        ax_kws:dict = None,
        show:bool = True,
        **box_kws,
)->Tuple[Union[plt.Axes, List[plt.Axes]], Union[List[dict], dict]]:
    """
    Draws the box and whiker plot

    parameters
    ----------
    data :
        array like (list, numpy array, pandas dataframe/series) or list of
        array likes. If list of array likes, the length of arrays in the list
        can be different.
    line_color :
        name of color/colors/cmap for lines/boundaries of box
    line_width :
        width of the box lines.
    fill_color :
        name of color/colors/cmap to fill the boxes. It can be any valid
         matplotlib color or cmap.
    labels : str/list (default=None)
        used for ticklabels of x-axes
    share_axes : bool (default=True)
        whether to draw all the histograms on one axes or not
    figsize : tuple (default=None)
        figure size as tuple (width, height)
    ax : plt.Axes, optional (default=None)
        matploltib axes on which to draw the plot
    ax_kws : dict (default=None)
        keyword arguments of :py:func:`easy_mpl.utils.process_axes`
    show : bool (default=show)
        whether to show the plot or not
    **box_kws :
        any additional keyword argument for :obj:`matplotlib.axes.Axes.boxplot`

    Returns
    -------
    tuple
        a tuple of two
            - plt.Axes or list of :obj:`matplotlib.axes`
            - a dictionary or list of dictionaries which consists of boxes,
              medians, whiskers, fliers

    Examples
    ---------
    >>> from easy_mpl import boxplot
    >>> boxplot(np.random.random((100, 5)))
    we can also provide arrays of different lengths
    >>> boxplot([np.random.random(100), np.random.random(90)])
    the color can be given as either color name or colormap
    >>> boxplot(np.random.random((100, 3)), fill_color=['pink', 'lightblue', 'lightgreen'])
    >>> boxplot(np.random.random((100, 3)), fill_color="viridis")

    See :ref:`sphx_glr_auto_examples_boxplot.py` for more examples

    """

    if ax is None:
        ax = plt.gca()

    _box_kws = {}

    data, labels = _unpack_data(data, labels, share_axes)

    if share_axes:
        nplots = 1
    else:
        nplots = len(labels)

    if box_kws is None:
        box_kws = dict()

    _box_kws.update(box_kws)

    f, axes = create_subplots(nplots, ax=ax, figsize=figsize)

    if isinstance(axes, np.ndarray):
        axes = axes.flatten().tolist()
    elif isinstance(axes, plt.Axes):
        axes = [axes]

    nboxes = len(labels)
    if len(labels)==1:
        nboxes = len(labels[0])
    fill_colors = _unpack_colors(fill_color, nboxes, share_axes)
    line_colors = _unpack_colors(line_color, nboxes, share_axes)
    line_widths = _unpack_linewidth(line_width, nboxes, share_axes)

    box_outs = []
    for (idx, name), x, ax in zip(enumerate(labels), data, axes):

        # in version 3.2.. giving DataFrame to ax.boxplot makes boxes for each row
        # in version 3.3.. giving DataFrame to ax.boxplot tries to make boxp for first row (columns)
        if is_dataframe(x) and matplotlib.__version__ < "3.3.0":
            x = x.values

        box_out = ax.boxplot(x, **_box_kws)
        box_outs.append(box_out)

        _set_box_props(fill_colors[idx], line_colors[idx],
                       line_widths[idx], box_out)

        _set_ticklabels(ax, share_axes, name, _box_kws)

        if ax_kws:
            process_axes(ax, **ax_kws)

    if show:
        plt.show()

    if len(box_outs)==1:
        box_outs = box_outs[0]

    if len(axes)==1:
        axes = axes[0]

    return axes, box_outs


def _set_ticklabels(ax, share_axes, name, box_kws):

    if name is not None:
        kws = dict()

        if share_axes:
            if box_kws.get('vert', True):
                ax.set_xticklabels(name)
            else:
                ax.set_yticklabels(name)
        else:
            if isinstance(name, (str, int)):
                if box_kws.get('vert', True):
                    ax.set_xticklabels([name])
                else:
                    ax.set_yticklabels([name], rotation=90, va='center')
            elif isinstance(name, list):
                if box_kws.get('vert', True):
                    ax.set_xticklabels(name)
                else:
                    ax.set_yticklabels(name, rotation=90, va='center')

        if share_axes and len(name) > 7:
            kws['rotation'] = 90
            ax.xaxis.set_tick_params(rotation=90)

    return


def _unpack_linewidth(line_width, nboxes, share_axes):
    if isinstance(line_width, (float, int)):
        line_widths = [[line_width] for _ in range(nboxes)]
    elif line_width is None:
        line_widths = [[None] for _ in range(nboxes)]

    if share_axes:
       line_widths = [[line_width[0] for line_width in line_widths]]

    return line_widths


def _unpack_colors(color, nboxes, share_axes)->list:
    if isinstance(color, str):
        if color in plt.colormaps():
            colors = make_cols_from_cmap(color, nboxes)
            colors = [[color] for color in colors]
        else:
            colors = [[color] for _ in range(nboxes)]
    elif is_rgb(color):
        colors = [[color] for _ in range(nboxes)]
    elif hasattr(color, '__len__'):
        assert len(color) == nboxes, f"{len(color)} colors for {nboxes} boxes?"
        colors = [[clr] for clr in color]
    elif color is None:
        colors = [[None] for _ in range(nboxes)]
    else:
        raise ValueError(f"{color} is not recognized as valid color")

    if share_axes:
       colors = [[color[0] for color in colors]]

    return colors


def _set_box_props(fill_color:list,
                   line_color:list,
                   line_width:list,
                   box_out):

    for idx, patch in enumerate(box_out['boxes']):
        if hasattr(patch, 'set_facecolor'):
            if fill_color[idx] is not None:
                patch.set_facecolor(fill_color[idx])
        elif hasattr(patch, 'set_markerfacecolor'):
            if fill_color[idx] is not None:
                patch.set_markerfacecolor(fill_color[idx])

        if hasattr(patch, 'set_color') and line_color[idx] is not None:
            patch.set_color(line_color[idx])

        if hasattr(patch, 'set_linewidth') and line_width[idx] is not None:
            patch.set_linewidth(line_width[idx])

    return


def _unpack_data(x, labels, share_axes:bool)->Tuple[list, list]:

    if isinstance(x, np.ndarray):
        if len(x) == x.size:
            X = [x]
            names = [[None]]
        else:
            if share_axes:
                X = [x]
                names = [[f"{i}" for i in range(x.shape[1])]]
            else:
                X = [x[:, i] for i in range(x.shape[1])]
                names = [f"{i}" for i in range(x.shape[1])]

    elif is_dataframe(x):
        if share_axes:
            names = [x.columns.tolist()]
            X = [x]
        else:
            X = []
            for col in x.columns:
                X.append(x[col].values)
            names = x.columns.tolist()

    elif is_series(x):
        X = [x.values]
        names = [[x.name]]

    elif isinstance(x, (list, tuple)) and isinstance(x[0], (list, tuple, np.ndarray)):
        assert all([len(array)==array.size for array in x]), f"""
        All arrays must be one dimensional."""
        X = [np.array(x_).reshape(-1,) for x_ in x]
        names = [None] * len(X)
        if share_axes:
            X = [X]
            names = [names]

    elif isinstance(x, (list, tuple)) and is_series(x[0]):  # list of series
        if share_axes:
            X = [x]
        else:
            X = x
        names = [x_.name for x_ in x]

    elif isinstance(x, (list, tuple)) and not is_dataframe(x[0]):
        X = [x]
        names = [None]
    else:
        raise ValueError(f"unrecognized type of x {type(x)}")

    if labels is not None:
        if isinstance(labels, str):
            labels = [labels]
        if share_axes:
            labels = [labels]
        #assert len(labels) == len(names), f"{len(names)} does not match data"
        names = labels

    return X, names




from ._main import plot
from ._pie import pie
from ._bar import bar_chart
from ._regplot import regplot
from ._imshow import imshow
from ._hist import hist
from ._scatter import scatter
from ._contour import contour
from ._dumbell import dumbbell_plot
from ._ridge import ridge
from ._parallel_coords import parallel_coordinates
from ._taylor_plot import taylor_plot
from ._lollipop import lollipop_plot
from ._circular_bar_plot import circular_bar_plot
from ._spider import spider_plot
from ._box import boxplot
from ._violin import violin_plot


from .utils import process_axes


__version__ = "0.21.3"




__all__ = ["bar_chart"]

import random
from typing import Union, List

import numpy as np
import matplotlib.pyplot as plt
from collections.abc import KeysView, ValuesView

from .utils import is_rgb
from .utils import BAR_CMAPS
from .utils import process_axes
from .utils import create_subplots
from .utils import to_1d_array, map_array_to_cmap


def bar_chart(
        values,
        labels=None,
        orient:str = 'h',
        sort:bool = False,
        max_bars:int = None,
        errors = None,
        color=None,
        cmap: Union[str, List[str]] = None,
        rotation:int = 0,
        bar_labels: Union[list, np.ndarray] = None,
        bar_label_kws=None,
        share_axes: bool = True,
        width = None,
        ax:plt.Axes = None,
        ax_kws: dict = None,
        show:bool = True,
        **kwargs
) -> Union[plt.Axes, List[plt.Axes]]:
    """
    plots bar chart

    Parameters
    -----------
        values : array like
            array like e.g. list/numpy array/ pandas series/ pandas dataframe / tuple
        labels : list, optional
            used for labeling each bar
        orient : `str`, optional
            orientation of bars. either 'h' or 'v'
        sort : bool, optional (default=None)
            whether to sort the bars based upon their values or not
        max_bars: int, optional (default=None)
            maximum number of bars to show
        errors : list, optional
            for error bars
        color : bool, optional (default=None)
            color for bars. It can any color value valid for matplotlib.
        cmap : str, optional (default=None)
            matplotlib colormap
        rotation : int, optional
            rotation angle of ticklabels
        bar_labels : list
            labels of the bars
        bar_label_kws : dict
            keyword arguments for :obj:`matplotlib.axes.Axes.bar_label`
        share_axes : bool (default=True)
            only relevant if given array is 2 dimentional. In such a case,
            this refers to whether to draw two bar charts on same axes or on two
            separate axes
        width : float
            width of the bars
        ax : :obj:`matplotlib.axes`, optional
            If not given, current available axes will be used
        ax_kws : dict, optional
            any keyword arguments for processing of axes that will go to
            :py:func:`easy_mpl.utils.process_axes`
        show : bool, optional
            whether to show the plot or not
        **kwargs :
            any additional keyword arguments for :obj:`matplotlib.axes.Axes.bar`
            or :obj:`matplotlib.axes.Axes.barh`

    Returns
    --------
    :obj:`matplotlib.axes`
        matplotlib Axes  or list of matplotlib Axes on which the bar_chart is drawn.
        If ``show`` is False, this axes can be used for further processing

    Examples
    --------
        >>> from easy_mpl import bar_chart
        >>> bar_chart([1,2,3,4,4,5,3,2,5])
        specifying labels
        >>> bar_chart([3,4,2,5,10], ['a', 'b', 'c', 'd', 'e'])
        sorting the data
        >>> bar_chart([1,2,3,4,4,5,3,2,5], sort=True)
        multiple bar charts
        >>> bar_chart(np.random.randint(0, 10, (5, 2)), color=['salmon', 'cadetblue'])

    See :ref:`sphx_glr_auto_examples_bar_chart.py` for more examples

    """

    if labels is None:
        if hasattr(values, "index") and hasattr("values", "name"):
            labels = values.index

    naxes = 1
    ncharts = 1
    if is_1d(values):
        values = to_1d_array(values)
    else:
        values = np.array(values)
        ncharts = values.shape[1]
        if share_axes:
            kwargs['edgecolor'] = kwargs.get('edgecolor', 'k')
        else:
            naxes = values.shape[1]

    colors = get_color(cmap, color, ncharts, values)

    figsize = None
    if 'figsize' in kwargs:
        figsize = kwargs.pop('figsize')

    if sort:
        assert ncharts == 1, f"""
        sorting is not allowed for more than 1 charts. ncharts are {ncharts}"""

    ax = maybe_create_axes(ax, naxes, figsize=figsize)

    if ncharts == 1:
        values, labels, bar_labels, colors = preprocess(values, labels,
                                                bar_labels, sort, max_bars, colors[0])
        ind = np.arange(len(values))
        bar_on_axes(ax[0], orient=orient, ax_kws=ax_kws, ind=ind,
                    values=values,
                    width=width, ticks=ind, labels=labels, color=colors,
                    bar_labels=bar_labels,
                   rotation=rotation, errors=errors,
                    bar_label_kws=bar_label_kws, kwargs=kwargs)

    elif share_axes:
        ind = np.arange(len(values))  # the label locations
        width = width or 1/ncharts * 0.9  # the width of the bars

        inds = []
        for idx in range(ncharts):
            if idx>0:
                ind = ind + width
            inds.append(ind)
        inds = np.column_stack(inds)
        ticks = np.mean(inds, axis=1)

        for idx in range(ncharts):

            _kwargs =kwargs.copy()
            _kwargs['label'] = _kwargs.get('label', idx)

            vals, labels, bar_labels, color = preprocess(values[:, idx], labels,
                                                  bar_labels, sort, max_bars, colors[idx])
            bar_on_axes(ax[0], orient, ax_kws,
                        inds[:, idx], vals, width, ticks, labels,
                        color, bar_labels,
                       rotation, errors, bar_label_kws, _kwargs)

    else:
        for idx in range(naxes):
            axes = ax[idx]
            data = values[:, idx]
            data, labels, bar_labels, color = preprocess(data, labels, bar_labels,
                                                  sort, max_bars, colors[idx])

            _kwargs = kwargs.copy()
            _kwargs['label'] = _kwargs.get('label', idx)

            ind = np.arange(len(data))
            bar_on_axes(axes, orient, ax_kws,
                        ind, data, width, ind, labels,
                        color, bar_labels,
                        rotation, errors,
                        bar_label_kws=bar_label_kws, kwargs=_kwargs)

    if show:
        plt.show()

    if len(ax) == 1:
        ax = ax[0]

    return ax


def maybe_create_axes(ax, naxes:int, figsize=None)->List[plt.Axes]:
    if ax is None:
        ax = plt.gca()
        if naxes>1:
            f, ax = create_subplots(ax=ax, naxes=naxes, figsize=figsize)
            ax = ax.flatten()
        else:
            if figsize:
                ax.figure.set_size_inches(figsize)
            ax = [ax]
    elif naxes>1:
        f, ax = create_subplots(ax=ax, naxes=naxes, figsize=figsize)
        ax = ax.flatten()
    else:
        if figsize:
            ax.figure.set_size_inches(figsize)
        ax = [ax]

    return ax


def handle_sort(sort, values, labels, bar_labels, color):
    if sort:
        sort_idx = np.argsort(values)
        values = values[sort_idx]
        labels = np.array(labels)[sort_idx]
        if bar_labels is not None:
            bar_labels = np.array(bar_labels)
            bar_labels = bar_labels[sort_idx]
            if 'float' in bar_labels.dtype.name:
                bar_labels = np.round(bar_labels, decimals=2)

        if isinstance(color, (list, np.ndarray, tuple)):
            if is_rgb(color[0]) or isinstance(color[0], str):
                color = np.array(color)[sort_idx]

    return values, labels, bar_labels, color


def handle_maxbars(max_bars, values, labels):
    if max_bars:
        n = len(values) - max_bars
        last_val = sum(values[0:-max_bars])
        values = values[-max_bars:]
        labels = labels[-max_bars:]
        values = np.append(last_val, values)
        labels = np.append(f"Rest of {n}", labels)
    return values, labels


def preprocess(data, labels, bar_labels, sort, max_bars, colors):
    if labels is None:
        labels = [f"F{i}" for i in range(len(data))]

    data, labels, bar_labels, colors = handle_sort(sort,
                                                   data,
                                                   labels,
                                                   bar_labels,
                                                   colors)

    data, labels = handle_maxbars(max_bars, data, labels)

    return data, labels, bar_labels, colors


def bar_on_axes(ax, orient, ax_kws, *args, **kwargs):
    if orient in ['h', 'horizontal']:
        horizontal_bar(ax, *args, **kwargs)
    else:
        vertical_bar(ax, *args, **kwargs)

    if ax_kws:
        process_axes(ax, **ax_kws)

    return


def horizontal_bar(ax, ind, values, width, ticks, labels, color, bar_labels,
                   rotation, errors, bar_label_kws, kwargs):

    if width:
        bar = ax.barh(ind, values, width, color=color, **kwargs)
    else:
        bar = ax.barh(ind, values, color=color, **kwargs)
    ax.set_yticks(ticks)
    ax.set_yticklabels(labels, rotation=rotation)

    set_bar_labels(bar, ax, bar_labels, bar_label_kws, errors,
                   values, ind)

    if 'label' in kwargs:
        ax.legend()
    return


def vertical_bar(ax, ind, values, width, ticks, labels, color, bar_labels,
                 rotation, errors, bar_label_kws, kwargs):

    bar = ax.bar(ind, values, width=width or 0.8, color=color, **kwargs)
    ax.set_xticks(ticks)
    ax.set_xticklabels(labels, rotation=rotation)

    set_bar_labels(bar, ax, bar_labels, bar_label_kws, errors,
                   ind, values)
    return


def set_bar_labels(bar, ax, bar_labels, bar_label_kws, errors,
                   values, ind):
    if bar_labels is not None:
        bar_label_kws = bar_label_kws or {'label_type': 'center'}
        if hasattr(ax, 'bar_label'):
            ax.bar_label(bar, labels=bar_labels, **bar_label_kws)
        else:
            bar.set_label(bar_labels)

    if errors is not None:
        ax.errorbar(values, ind, xerr=errors, fmt=".",
                    color="black")
    return


def is_1d(array):
    if isinstance(array, (KeysView, ValuesView)):
        array = np.array(list(array))
    else:
        array = np.array(array)
    if len(array)==array.size:
        return True
    return False


def get_color(cmap, color, ncharts, data)->list:

    data = np.reshape(data, (len(data), -1))

    if not isinstance(cmap, list):
        cmap = [cmap for _ in range(ncharts)]

    if not isinstance(color, list):
        color = [color for _ in range(ncharts)]
    elif ncharts == 1:
        # the user has specified separate color for each bar
        # in next for loop we don't want to get just firs color from the list
        color = [color]

    colors = []
    for idx in range(ncharts):

        _cmap = cmap[idx] or random.choice(BAR_CMAPS)
        cm, _ = map_array_to_cmap(data[:, idx], _cmap)

        clr = color[idx] if color[idx] is not None else cm

        colors.append(clr)

    return colors





__all__ = ["pie"]

from typing import Union

import numpy as np
import matplotlib.pyplot as plt

from easy_mpl.utils import process_axes


def pie(
        vals: Union[list, np.ndarray] = None,
        fractions: Union[list, np.ndarray] = None,
        labels: list = None,
        autopct = '%1.1f%%',
        ax: plt.Axes = None,
        ax_kws: dict = None,
        show: bool = True,
        **kwargs
) -> tuple:
    """
    draws the pie chart

    Parameters
    ----------
        vals : array like,
            unique values and their counts will be inferred from this array.
        fractions : list, array, optional
            if given, vals must not be given
        labels : list, array, optional
            labels for unique values in vals, if given, must be equal to unique vals
            in vals. Otherwise "unique_value (counts)" will be used for labeling.
        autopct :
        ax : plt.Axes, optional
            the :obj:`matplotlib.axes` on which to draw, if not given current active axes will be used
        ax_kws : dict, optional
            keyword arguments for :py:func:`easy_mpl.utils.process_axes`
        show: bool, optional (default=True)
            whether to show the plot or not
        **kwargs: optional
            any keyword argument will go to :obj:`matplotlib.axes.Axes.pie`

    Returns
    -------
    outs
        same what is returned by :obj:`matplotlib.axes.Axes.pie`

    Example
    -------
        >>> import numpy as np
        >>> from easy_mpl import pie
        >>> pie(np.random.randint(0, 3, 100))
        or by directly providing fractions
        >>> pie([0.2, 0.3, 0.1, 0.4])
        ... # to explode 0.3
        >>> explode = (0, 0.1, 0, 0, 0)
        >>> pie(fractions=[0.2, 0.3, 0.15, 0.25, 0.1], explode=explode)

    See :ref:`sphx_glr_auto_examples_pie.py` for more examples

    .. _axes.pie:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.pie.html
    """
    # todo, add example for and partial pie chart

    if ax_kws is None:
        ax_kws = dict()

    if ax is None:
        ax = plt.gca()
        if 'figsize' in ax_kws:
            figsize = ax_kws.pop('figsize')
            ax.figure.set_size_inches(figsize)

    if fractions is None:
        uniques, counts = np.unique(vals, return_counts=True)
        fractions = counts / counts.sum()
        vals = {k:v for k,v in zip(uniques, counts)}

        if labels is None:
            labels = [f"{value} ({count}) " for value, count in vals.items()]
    else:
        assert vals is None

    outs = ax.pie(fractions,
                  labels=labels,
                  autopct=autopct,
           **kwargs)

    ax.axis('equal')  # Equal aspect ratio ensures that pie is drawn as a circle.

    if ax_kws:
        process_axes(ax, **ax_kws)

    if show:
        plt.show()

    return outs
